
import pandas as pd
from collections import namedtuple
import pyarrow
from tqdm.auto import tqdm
from ..common.api_request import make_request

def list_rows(
    *, uri, type="tuple", max_results=None, selected_variables=None, mapped_variables=None
):
    format = 'arrow'
    read_session = make_request(
        method="post",
        path=f'{uri}/readSessions',
        parse_response=True,
        payload={
            "selectedVariables": selected_variables,
            "maxResults": max_results,
            "format": format
        },
    )

    progressbar = tqdm(total=max_results, leave=False)
    stream_results = []
    for stream in read_session["streams"]:
        arrow_response = make_request(
            method="get",
            path=f'/readStreams/{stream["id"]}',
            stream=True,
            parse_response=False,
        )

        reader = pyarrow.ipc.open_stream(arrow_response.raw)
        batches = []
        for batch in reader:
            batches.append(batch)
            progressbar.update(batch.num_rows)

        if type == "tuple":
            stream_results.append(pyarrow.Table.from_batches(batches).to_pydict())
        else:
            stream_results.append(pyarrow.Table.from_batches(batches).to_pandas())

    if type == "tuple":
        Row = namedtuple(
            "Row",
            [variable["name"] for variable in mapped_variables],
        )
        res = []
        for pydict in stream_results:
            keys = list(pydict.keys())
            for i in range(len(pydict[keys[0]])):
                if len(res) == max_results:
                    break
                res.append(Row(*[format_tuple_type(pydict[variable["name"]][i], variable["type"]) if variable["name"] in pydict else None for variable in mapped_variables]))

        progressbar.close()
        return res
    else:
        df = pd.concat(stream_results) if len(stream_results) > 1 else stream_results[0]
        df = set_dataframe_types(df, mapped_variables)
        if len(df.index) > max_results:
            return df.iloc[0:max_results, :]

        progressbar.close()
        return df

    # Old avro parsing, removed in preference of Arrow
    # if type == "tuple":
    #     res = []
    #     for stream in read_session["streams"]:
    #         avro_response = make_request(
    #             method="get",
    #             path=f'/readStreams/{stream["id"]}',
    #             stream=True,
    #             parse_response=False,
    #         )
    #         parsed_schema = fastavro.parse_schema(read_session["schemas"][stream["schemaIndex"]])
    #
    #         Row = namedtuple(
    #             "Row",
    #             [variable["name"] for variable in mapped_variables],
    #         )
    #         while True:
    #             try:
    #                 res.append(Row(**fastavro.schemaless_reader(avro_response.raw, parsed_schema)))
    #             except Exception as err:
    #                 break
    #             progressbar.update()
    #
    #         progressbar.close()
    #
    #     return res
    # else:
    #     res = []
    #     for stream in read_session["streams"]:
    #         avro_response = make_request(
    #             method="get",
    #             path=f'/readStreams/{stream["id"]}',
    #             stream=True,
    #             parse_response=False,
    #         )
    #         parsed_schema = fastavro.parse_schema(read_session["schemas"][stream["schemaIndex"]])
    #
    #         while True:
    #             try:
    #                 res.append(fastavro.schemaless_reader(avro_response.raw, parsed_schema))
    #             except Exception as err:
    #                 break
    #             progressbar.update()
    #
    #         progressbar.close()
    #
    #     df = pd.DataFrame(res, dtype="string")
    #     return set_dataframe_types(df, mapped_variables)


def format_tuple_type(val, type):
    if val is None:
        return val
    elif type == "integer":
        return int(val)
    elif type == "float":
        return float(val)
    elif type == "date":
        return str(val)
    elif type == "dateTime":
        return str(val)
    elif type == "time":
        return str(val)
    elif type == "boolean":
        return bool(val)
    else:
        return str(val)


def set_dataframe_types(df, variables):
    for variable in variables:
        name = variable["name"]
        type = variable["type"]

        # TODO: need to finalize what types we're returning
        if type == "integer":
            df[name] = pd.to_numeric(df[name])
            # df[name] = df[name].astype("Int64")
        elif type == "float":
            df[name] = df[name].astype("float64")
            # df[name] = df[name].astype("Float64")
        elif type == "date":
            df[name] = pd.to_datetime(df[name], errors="coerce")
        elif type == "dateTime":
            df[name] = pd.to_datetime(df[name], errors="coerce")
        elif type == "time":
            df[name] = pd.to_timedelta(df[name], errors="coerce")
        elif type == "boolean":
            df[name].replace(
                to_replace=["TRUE", "FALSE"], value=[True, False], inplace=True
            )
            # Pandas seems to throw errors if all the boolean values are NA, in which case we should just ignore and fall back to a string dtype
            df[name] = df[name].astype("boolean", errors="ignore")
        else:
            df[name] = df[name].astype("string")

    return df



import warnings
from .Dataset import Dataset
from urllib.parse import quote as quote_uri
from ..common.api_request import make_paginated_request

class Organization:
    def __init__(self, name):
        self.name = name
        self.uri = f"/organizations/{quote_uri(self.name)}"

    def dataset(self, name, *, version="current"):
        return Dataset(name, organization=self, version=version)

    def list_datasets(self, max_results=None):
        datasets = make_paginated_request(
            path=f"{self.uri}/datasets", page_size=100, max_results=max_results
        )
        return [
            Dataset(dataset["name"], organization=self, properties=dataset) for dataset in datasets
        ]



from ..common.api_request import make_request
import json
import os
import time
from ..common.list_rows import list_rows
import warnings


class Query:
    def __init__(
        self,
        query,
        *,
        default_project=None,
        default_dataset=None,
    ):
        if not default_project and not default_dataset:
            if os.getenv("REDIVIS_DEFAULT_PROJECT"):
                default_project = os.getenv("REDIVIS_DEFAULT_PROJECT")
            elif os.getenv("REDIVIS_DEFAULT_DATASET"):
                default_dataset = os.getenv("REDIVIS_DEFAULT_DATASET")

        self.properties = make_request(
            method="post",
            path="/queries",
            payload={
                "query": query,
                "defaultProject": default_project if default_project else None,
                "defaultDataset": default_dataset if default_dataset else None,
            },
        )
        self.uri = f"/queries/{self.properties['id']}"

    def __getitem__(self, key):
        return (
            self.properties[key] if self.properties and key in self.properties else None
        )

    def __str__(self):
        return json.dumps(self.properties, indent=2)

    def get(self):
        self.properties = make_request(method="GET", path=self.uri)
        return self

    def list_rows(self, max_results=None, *, limit=None):
        if limit and max_results is None:
            warnings.warn(
                "The limit parameter has been renamed to max_results, and will be removed in a future version of this library",
                DeprecationWarning,
            )
            max_results = limit

        self._wait_for_finish()
        variables = self.properties["outputSchema"]

        max_results = (
            min(max_results, int(self.properties["outputNumRows"]))
            if max_results is not None
            else self.properties["outputNumRows"]
        )

        return list_rows(
            uri=self.uri,
            max_results=max_results,
            mapped_variables=variables,
            type="tuple",
        )

    def to_dataframe(self, max_results=None, *, limit=None):
        if limit and max_results is None:
            warnings.warn(
                "The limit parameter has been renamed to max_results, and will be removed in a future version of this library",
                DeprecationWarning,
            )
            max_results = limit

        self._wait_for_finish()
        variables = self.properties["outputSchema"]

        max_results = (
            min(max_results, int(self.properties["outputNumRows"]))
            if max_results is not None
            else self.properties["outputNumRows"]
        )

        return list_rows(
            uri=self.uri,
            max_results=max_results,
            mapped_variables=variables,
            type="dataframe",
        )

    def _wait_for_finish(self):
        while True:
            if self.properties["status"] == "completed":
                break
            elif self.properties["status"] == "failed":
                raise Exception(
                    f"Query job failed with message: {self.properties['errorMessage']}"
                )
            elif self.properties["status"] == "cancelled":
                raise Exception(f"Query job was cancelled")
            else:
                time.sleep(2)
                self.get()



import json
from urllib.parse import quote as quote_uri
import warnings
import os

from .Upload import Upload
from .Variable import Variable
from ..common.list_rows import list_rows
from ..common.api_request import make_request, make_paginated_request


class Table:
    def __init__(
        self,
        name,
        *,
        sample=False,
        dataset=None,
        project=None,
        properties=None,
    ):
        dataset, project = get_table_parents(dataset, project)
        parent = dataset or project
        owner = parent.user or parent.organization
        sample_string = ":sample" if sample else ""
        version_string = f":{dataset.version}" if dataset else ""
        self.name = name
        self.dataset = dataset
        self.project = project
        self.identifier = (
            f"{owner.name}.{parent.name}{version_string}.{self.name}{sample_string}"
        )
        self.uri = f"/tables/{quote_uri(self.identifier)}"
        self.properties = properties

    def __getitem__(self, key):
        return (
            self.properties[key] if self.properties and key in self.properties else None
        )

    def __str__(self):
        return json.dumps(self.properties, indent=2)

    def create(self, *, description=None, upload_merge_strategy="append"):
        response = make_request(
            method="POST",
            path=f"{self.dataset.uri}/tables",
            payload={
                "name": self.name,
                "description": description,
                "uploadMergeStrategy": upload_merge_strategy,
            },
        )
        self.properties = response
        self.uri = self.properties["uri"]
        return self

    def delete(self):
        make_request(
            method="DELETE",
            path=self.uri,
        )
        return

    def get(self):
        self.properties = make_request(method="GET", path=self.uri)
        self.uri = self.properties["uri"]
        return self

    def exists(self):
        try:
            make_request(method="GET", path=self.uri)
            return True
        except Exception as err:
            if err.args[0]["status"] != 404:
                raise err
            return False

    def list_rows(self, max_results=None, *, limit=None, variables=None):
        if limit and max_results is None:
            warnings.warn(
                "The limit parameter has been renamed to max_results, and will be removed in a future version of this library",
                DeprecationWarning,
            )
            max_results = limit

        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        max_results = (
            min(max_results, int(self.properties["numRows"]))
            if max_results is not None
            else self.properties["numRows"]
        )

        mapped_variables = get_mapped_variables(variables, self.uri)

        return list_rows(
            uri=self.uri,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            type="tuple",
        )

    def list_uploads(self, *, max_results=None):
        uploads = make_paginated_request(
            path=f"{self.uri}/uploads", max_results=max_results
        )
        return [Upload(upload) for upload in uploads]

    def list_variables(self, *, max_results=None):
        variables = make_paginated_request(
            path=f"{self.uri}/variables", page_size=1000, max_results=max_results
        )
        return [
            Variable(variable["name"], table=self, properties=variable)
            for variable in variables
        ]

    def to_dataframe(self, max_results=None, *, limit=None, variables=None):
        if limit and max_results is None:
            warnings.warn(
                "The limit parameter has been renamed to max_results, and will be removed in a future version of this library",
                DeprecationWarning,
            )
            max_results = limit

        if not self.properties or not hasattr(self.properties, "numRows"):
            self.get()

        max_results = (
            min(max_results, int(self.properties["numRows"]))
            if max_results is not None
            else self.properties["numRows"]
        )

        mapped_variables = get_mapped_variables(variables, self.uri)
        return list_rows(
            uri=self.uri,
            max_results=max_results,
            selected_variables=variables,
            mapped_variables=mapped_variables,
            type="dataframe",
        )

    def update(self, *, name=None, description=None, upload_merge_strategy=None):
        payload = {}
        if name:
            payload["name"] = name
        if upload_merge_strategy:
            payload["uploadMergeStrategy"] = upload_merge_strategy
        if description is not None:
            payload["description"] = description

        response = make_request(
            method="PATCH",
            path=f"{self.uri}",
            payload=payload,
        )
        self.properties = response
        return

    def upload(
        self,
        name,
        *,
        type="delimited",
        schema=None,
        has_header_row=True,
        skip_bad_records=False,
        allow_quoted_newlines=False,
        quote_character='"',
        delimiter=None,
        #     This is deprecated
        data=None,
        remove_on_fail=True,
    ):
        upload = Upload(
            table=self,
            name=name,
        )
        if data is not None:
            warnings.warn(
                "Passing data directly to the upload constructor is deprecated. Please call table.upload('filename').create(data) instead.",
                DeprecationWarning,
            )
            upload.create(
                schema=schema,
                type=type,
                has_header_row=has_header_row,
                skip_bad_records=skip_bad_records,
                allow_quoted_newlines=allow_quoted_newlines,
                quote_character=quote_character,
                delimiter=delimiter,
            )
            upload.upload_file(
                data,
                wait_for_finish=True,
                raise_on_fail=True,
                remove_on_fail=remove_on_fail,
            )

        return upload

    def variable(self, name):
        return Variable(name, table=self)


def get_mapped_variables(variables, uri):
    all_variables = make_paginated_request(path=f"{uri}/variables", page_size=1000)

    if variables is None:
        return all_variables
    else:
        lower_variable_names = [variable.lower() for variable in variables]
        variables_list = list(
            filter(
                lambda variable: variable["name"].lower() in lower_variable_names,
                all_variables,
            )
        )
        variables_list.sort(
            key=lambda variable: lower_variable_names.index(variable["name"].lower())
        )
        return variables_list


def get_table_parents(dataset, project):
    from .User import User

    if dataset or project:
        return dataset, project
    elif os.getenv("REDIVIS_DEFAULT_PROJECT") is not None:
        return None, User(os.getenv("REDIVIS_DEFAULT_PROJECT").split(".")[0]).project(
            os.getenv("REDIVIS_DEFAULT_PROJECT").split(".")[1]
        )
    elif os.getenv("REDIVIS_DEFAULT_DATASET") is not None:
        return (
            User(os.getenv("REDIVIS_DEFAULT_DATASET").split(".")[0]).dataset(
                os.getenv("REDIVIS_DEFAULT_DATASET").split(".")[1]
            )
        ), None
    else:
        raise Exception(
            "Cannot reference an unqualified table if the neither the REDIVIS_DEFAULT_PROJECT or REDIVIS_DEFAULT_DATASET environment variables are set."
        )
